from __future__ import unicode_literals

import os
import configparser

def stdoe_folder():
    home = os.path.join(os.path.expanduser('~'))
    conf = configparser.ConfigParser()
    conf.read(os.path.join(home, '.config', 'lsf', 'config'))
    return conf.get('default', 'stdoe_folder')



from __future__ import unicode_literals

from subprocess import Popen
import os
import subprocess
import re
from .config import stdoe_folder
from limix_util.path import make_sure_path_exists
from limix_util.string import make_sure_unicode

_max_nfiles = 1000

def killall(force=False):
    if force:
        subprocess.call("bkill -r -u horta 0", shell=True)
    else:
        subprocess.call("bkill -u horta 0", shell=True)

def get_output_files(i, runid):
    pr = str(int(i / _max_nfiles))
    base = os.path.join(stdoe_folder(), runid, pr)
    make_sure_path_exists(base)
    ofile = os.path.join(base, 'out_%d.txt' % i)
    efile = os.path.join(base, 'err_%d.txt' % i)
    return (ofile, efile)

_stats = [None]
def get_jobs_stat():
    if _stats[0] is None:
        cmd = 'bjobs -a -o "JOBID STAT" -noheader'
        r = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT,
                                    universal_newlines=True)
        r = make_sure_unicode(r)
        r = r.strip()
        if r == 'No job found':
            _stats[0] = {}
        else:
            _stats[0] = {int(row.split(' ')[0]):row.split(' ')[1] for row in r.split('\n')}
    return _stats[0]

def group_jobids(grp):
    procs = Popen("bjobs -g %s -w | awk '{print $1}'" % grp,
                  shell=True, stdout=subprocess.PIPE,
                  universal_newlines=True).stdout.read()
    procs = procs.split('\n')
    procs = procs[1:-1]
    return [int(p) for p in procs]

def kill_group(grp, block=True):
    jobids = group_jobids(grp)
    procs = []
    for jobid in jobids:
        p = Popen("bkill %d &> /dev/null" % jobid, shell=True,
                  universal_newlines=True)
        procs.append(p)

    if block:
        for p in procs:
            p.wait()

def _try_clean_runid(runid):
    c = re.compile(r'^.*(\d\d\d\d-\d\d-\d\d-\d\d-\d\d-\d\d).*$')
    m = c.match(runid)
    if m is None:
        return runid
    return m.group(1)

def proper_runid(what):
    m = re.match(r'^last(\^*)$', what)
    if m is None:
        return _try_clean_runid(what)
    n = len(m.group(1)) + 1
    runids = get_runids()
    runids.sort()
    return runids[max(-len(runids), -n)]

def get_runids():
    c = re.compile(r'^\d\d\d\d-\d\d-\d\d-\d\d-\d\d-\d\d$')
    files = os.listdir(config.stdoe_folder())
    runids = [f for f in files if c.match(f)]
    return runids




import re
from limix_util.string import make_sure_unicode
from io import StringIO
import humanfriendly as hf
from limix_util.pickle import SlotPickleMixin
from . import util


class Job(SlotPickleMixin):
    __slots__ = ['jobid', 'finished', '_exit_status', 'cmd', '_bcmd',
                 'odata', 'edata', 'runid', '_hasfinished', '_stat_cache',
                 'mkl_nthreads']

    def __init__(self, jobid, cmd):
        self.runid = None
        self.jobid = jobid
        self._exit_status = None
        self._hasfinished = False
        self.cmd = cmd
        self._bcmd = []
        self.odata = ''
        self.edata = ''
        self.mkl_nthreads = 1
        self._stat_cache = None

    def stdout(self):
        (fp, _) = util.get_output_files(self.jobid, self.runid)
        try:
            with open(fp, 'r') as f:
                return f.read()
        except IOError:
            return None
        return None

    def stderr(self):
        (_, fp) = util.get_output_files(self.jobid, self.runid)
        try:
            with open(fp, 'r') as f:
                return f.read()
        except IOError:
            return None
        return None

    @property
    def group(self):
        return '/cluster/%s' % self.runid

    @property
    def submission_status(self):
        if self.os_jobid:
            return 'success'
        return self.edata

    @property
    def os_jobid(self):
        m = re.match(r'^Job <(\d+)>.*$', make_sure_unicode(self.odata))
        if m:
            return int(m.group(1))
        return None

    @property
    def bcmd(self):
        return self._bcmd

    @bcmd.setter
    def bcmd(self, bc):
        self._bcmd = bc

    @property
    def full_cmd(self):
        n = self.mkl_nthreads
        middle  = ['env', 'MKL_NUM_THREADS=%d' % n]
        middle += ['MKL_DYNAMIC=TRUE']
        return self.bcmd + middle + self.cmd

    def stat(self):
        if self._stat_cache:
            return self._stat_cache

        stats = util.get_jobs_stat()

        if self.os_jobid not in stats:
            if self.hassubmitted():
                self._stat_cache = 'DONE_OR_EXIT'
                return self._stat_cache
            else:
                return 'UNKNOWN'

        r = stats[self.os_jobid]
        if r == 'DONE' or r == 'EXIT':
            self._stat_cache = r
        return r

    def hassubmitted(self):
        return ("Job <%d> " % self.os_jobid + "is submitted") in make_sure_unicode(self.odata)

    def ispending(self):
        return self.stat() == 'PEND'

    def isrunning(self):
        return self.stat() == 'RUN'

    def hasfinished(self):
        if self._hasfinished:
            return True
        self._hasfinished = (self.stat() == 'DONE' or self.stat() == 'EXIT' or
                             self.stat() == 'DONE_OR_EXIT')
        return self._hasfinished

    def exit_status(self):
        if not self.hasfinished():
            return None

        if self._exit_status is not None:
            return self._exit_status

        data = self.stdout()
        if data is None:
            return None

        buf = StringIO(data)
        found = False
        for line in buf:
            if line.strip() == 'Your job looked like:':
                found = True
                break

        if found:
            buf.readline()
            buf.readline()
            buf.readline()
            buf.readline()
            buf.readline()
            buf.readline()
            status_msg = buf.readline().strip()
            if status_msg == 'Successfully completed.':
                self._exit_status = 0
            else:
                m = re.match(r'^Exited with exit code (\d+)\.$', status_msg)
                if m is None:
                    self._exit_status = -1
                else:
                    self._exit_status = int(m.group(1))
        else:
            self._exit_status = None

        return self._exit_status

    def resource_info(self):
        if not self.hasfinished():
            return None
        data = self.stdout()

        if data is None:
            return None

        buf = StringIO(data)
        found = False
        for line in buf:
            if line.strip() == 'Resource usage summary:':
                found = True
                break

        if found:
            buf.readline()
            cpu_time = buf.readline()
            max_memory = buf.readline()
            m = re.match(r'^.*Max Memory : [^\d]*(\d.*)$', max_memory)
            if m is None:
                max_memory = None
            else:
                max_memory = hf.parse_size(m.group(1))
            avg_memory = buf.readline()
            req_memory = buf.readline()
            m = re.match(r'^.*Total Requested Memory : [^\d]*(\d.*)$', req_memory)
            if m is None:
                req_memory = None
            else:
                req_memory = hf.parse_size(m.group(1))
            delta_memory = buf.readline()
            return dict(max_memory=max_memory, req_memory=req_memory)

        return None

    #
    # def stdout(self):
    #     ofile, _ = _get_output_files(self.jobid, self.runid)
    #     try:
    #         with open(ofile, 'r') as f:
    #             return f.read()
    #     except IOError:
    #         return ""
    #
    # def stderr(self):
    #     _, efile = _get_output_files(self.jobid, self.runid)
    #     try:
    #         with open(efile, 'r') as f:
    #             return f.read()



from __future__ import unicode_literals

from concurrent.futures import ProcessPoolExecutor
import logging
import os
from os.path import join
from copy import copy
import re
import atexit
import subprocess
from subprocess import list2cmdline
import humanfriendly as hf
from limix_util.report import BeginEnd, ProgressBar
import limix_util.pickle as pickle_
from limix_util.string import make_sure_unicode
from limix_util.path import make_sure_path_exists
from limix_util.path import touch
from . import config
from . import util
from . import job

_cluster_runs = dict()
def load(runid):
    if runid not in _cluster_runs:
        folder = join(config.stdoe_folder(), runid)
        if os.path.exists(join(folder, '.deleted')):
            return None
        cr = pickle_.unpickle(join(folder, 'cluster_run.pkl'))
        _cluster_runs[runid] = cr
    return _cluster_runs[runid]

def rm(runid):
    folder = join(config.stdoe_folder(), runid)
    touch(join(folder, '.deleted'))

def exists(runid):
    if runid not in _cluster_runs:
        return os.path.exists(join(config.stdoe_folder(), runid,
                                   'cluster_run.pkl'))
    return True

def _submit_job(job):
    fcmd = job.full_cmd
    p = subprocess.Popen(list2cmdline(fcmd), shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         universal_newlines=True)
    (job.odata, job.edata) = p.communicate()
    job.odata = make_sure_unicode(job.odata)
    job.edata = make_sure_unicode(job.edata)
    p.stdout.close()
    p.stderr.close()
    return job

class ClusterRunBase(pickle_.SlotPickleMixin):
    __slots__ = ['requests', 'megabytes', 'jobs', 'group', 'runid', 'title',
                 'nprocs', 'mkl_nthreads', 'queue']
    def __init__(self, title):
        self.requests = []
        self.megabytes = 4096
        self.nprocs = 1
        self.jobs = []
        self.runid = None
        self.title = title
        self.mkl_nthreads = 1
        self.queue = None

class ClusterRun(ClusterRunBase):
    __slots__ = []
    def __init__(self, title='notitle'):
        super(ClusterRun, self).__init__(title)

    def store(self):
        with BeginEnd('Storing cluster commands', silent=True):
            pickle_.pickle(self, join(config.stdoe_folder(), self.runid,
                           'cluster_run.pkl'))

    def request(self, req):
        if self.runid is not None:
            raise Exception('This command set has already been sent.')

        if req == 'panfs':
            self.requests.append('panfs_nobackup_production')
        elif req == 'fasttmp':
            self.requests.append('fasttmp')
        elif req == 'gpfs':
            self.requests.append('gpfs')
        else:
            raise Exception("Don't know about %s." % req)

    def kill(self, block=True):
        grp = '/cluster/%s' % self.runid
        util.kill_group(grp, block=block)


    @property
    def memory(self):
        nbytes = int(round(self.megabytes / 1024. / 1024.))
        return hf.format_size(nbytes)

    @memory.setter
    def memory(self, siz):
        if self.runid is not None:
            raise Exception('This command set has already been sent.')

        nbytes = hf.parse_size(siz)
        self.megabytes = int(round(nbytes / 1024. / 1024.))

    def add(self, cmd):
        if self.runid is not None:
            raise Exception('This command set has already been sent.')

        cmd = cmd if isinstance(cmd, list) else [cmd]
        cmd = [str(c) for c in cmd]

        j = job.Job(len(self.jobs), cmd)
        self.jobs.append(j)

    def _dryrun_job(self, job):
        print(list2cmdline(job.full_cmd))

    # def _parallel_run_job(self, executor, job):
    #     fcmd = job.full_cmd
    #     f = executor.submit(subprocess.Popen, list2cmdline(fcmd), shell=True,
    #                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    #     # process = subprocess.Popen(list2cmdline(fcmd), shell=True,
    #     #                  stdout=subprocess.PIPE,
    #     #                  stderr=subprocess.PIPE)
    #     return f

    # def _run_job_wait(self, executor, job):
    #     fcmd = job.full_cmd
    #     f = executor.submit(subprocess.Popen, list2cmdline(fcmd), shell=True,
    #                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def _sequential_run_job(self, job):
        os.system(list2cmdline(job.cmd))

    def resubmit(self, jobid):
        raise NotImplementedError('me implementa ai porra')

    def run(self, dryrun=False):
        if self.runid is not None:
            raise Exception('This command set has already been sent.')

        self.runid = _generate_runid()
        if not dryrun:
            make_sure_path_exists(join(config.stdoe_folder(), self.runid))
        bcmd = self._init_parse()
        self._parse_requests(bcmd)
        self._parse_memory(bcmd)
        self._parse_nprocs(bcmd)
        bcmd += ['-g', '/cluster/%s' % self.runid]
        if self.queue:
            bcmd += ['-q', self.queue]
        self._end_parse(bcmd)

        ujobs = []
        max_workers = 500
        with BeginEnd('Submitting jobs'):
            pb = ProgressBar(len(self.jobs))
            i = 0
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                for job in executor.map(_submit_job, self.jobs):
                    job.runid = self.runid
                    ujobs.append(job)
                    i += 1
                    pb.update(i)
            pb.finish()

        self.jobs = ujobs
        print('   %d jobs have been submitted   '
              % len(self.jobs))

        print("Run ID: %s" % self.runid)

        self.store()

        return self.runid

    def _init_parse(self):
        return ['bsub']

    def _end_parse(self, bcmd):
        for job in self.jobs:
            ofile, efile = self._get_output_files(job.jobid)
            job.bcmd = copy(bcmd) + ['-o', ofile, '-e', efile]
            job.mkl_nthreads = self.mkl_nthreads

    def _get_output_files(self, jobid):
        _max_nfiles = 1000
        pr = str(int(jobid / _max_nfiles))
        base = join(config.stdoe_folder(), self.runid, pr)
        make_sure_path_exists(base)
        ofile = join(base, 'out_%d.txt' % jobid)
        efile = join(base, 'err_%d.txt' % jobid)
        return (ofile, efile)


    def _parse_requests(self, bcmd):
        bcmd.append("-R")
        bcmd.append("select[%s]" % (','.join(self.requests)))

    def _parse_memory(self, bcmd):
        m = self.megabytes
        bcmd.append("-M")
        bcmd.append("%d" % m)
        bcmd.append("-R")
        bcmd.append("rusage[mem=%d]" % m)

    def _parse_nprocs(self, bcmd):
        bcmd.append('-n')
        bcmd.append('%d' % self.nprocs)

    def __str__(self):
        njobs = len(self.jobs)
        if njobs == 0:
            return self.runid + ': no job'
        jsum = list2cmdline(self.jobs[0].cmd)
        return self.runid + ' (%d jobs): %s' % (njobs, jsum)

    @property
    def number_jobs(self):
        return len(self.jobs)

    @property
    def number_jobs_pending(self):
        return sum([1 for job in self.jobs if job.ispending()])

    @property
    def number_jobs_running(self):
        return sum([1 for job in self.jobs if job.isrunning()])

    @property
    def number_jobs_finished(self):
        return sum([1 for job in self.jobs if job.hasfinished()])

    @property
    def number_jobs_succeed(self):
        return sum([1 for job in self.jobs if job.hasfinished() and\
                                              job.exit_status() == 0])

    @property
    def number_jobs_failed(self):
        return sum([1 for job in self.jobs if job.hasfinished() and\
                                              job.exit_status() != 0])

def _generate_runid():
    from time import gmtime, strftime
    return strftime('%Y-%m-%d-%H-%M-%S', gmtime())

_registered_cluster_runs = dict()
def _register_cluster_run(cr):
    if id(cr) not in _registered_cluster_runs:
        _registered_cluster_runs[id(cr)] = cr

def _update_storage():
    for cr in list(_registered_cluster_runs.values()):
        cr.store()
atexit.register(_update_storage)

def get_bjob(runid, jobid):
    cr = load(runid)
    _register_cluster_run(cr)
    return cr.jobs[jobid]

def get_groups_summary(nlast=10):

    awk = "awk -F\" \" '{print $1, $2, $3, $4, $5, $6, $7}'"
    cmd = "bjgroup | grep -E \".*`whoami`$\" | %s" % awk
    msg = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    msg = msg.strip()
    lines = msg.split('\n')
    table = [line.split(' ') for line in lines]

    table = _clear_groups_summary(table)
    table.sort(key=lambda x: x[0], reverse=True)

    table_out = []
    logger = logging.getLogger(__file__)
    for row in table:
        runid = row[0].strip('/').split('/')[1]
        try:
            cr = load(runid)
            if cr is None:
                continue
        except ImportError as e:
            logger.warn('Could not load cluster run %s. Reason: %s.', runid,
                        str(e))
            row.append('UNK')
            row.append('UNK')
            row.append('UNK')
        except IOError as e:
            pass
        #     logger.warn('Could not load cluster run %s. Reason: %s.', runid,
        #                 str(e))
        #     row.append('UNK')
        #     row.append('UNK')
        #     row.append('UNK')
        else:
            try:
                row.append(cr.number_jobs_failed)
            except TypeError:
                row.append('UNK')

            try:
                row.append(cr.number_jobs_succeed)
            except TypeError:
                row.append('UNK')

            try:
                row.append(cr.title)
            except TypeError:
                row.append('UNK')

            table_out.insert(0, row)
            if len(table_out) >= nlast:
                break

    header = ['group_name', 'njobs', 'pend', 'run', 'ssusp', 'ususp', 'finish',
              'failed', 'succeed', 'title']
    table_out.insert(0, header)
    return table_out

def _clear_groups_summary(table):
    ntable = []
    for row in table:
        if not _isrun_group(row[0]):
            continue
        row[1:] = [int(c) for c in row[1:]]
        ntable.append(row)
    return ntable

def _isrun_group(name):
    name = name.strip('/')
    names = name.split('/')
    if len(names) != 2:
        return False
    return names[0] == 'cluster' and _isrunid(names[1])

_runid_matcher = re.compile(r'^\d\d\d\d-\d\d-\d\d-\d\d-\d\d-\d\d$')
def _isrunid(runid):
    return _runid_matcher.match(runid) is not None



from __future__ import unicode_literals

from argparse import ArgumentParser
from tabulate import tabulate
import logging
from . import clusterrun
from . import util
from . import config

def do_root(_):
    print(config.stdoe_folder())

def _do_run_status(runid):
    cc = clusterrun.load(runid)
    if cc is None:
        print('Runid %s could not be found.' % runid)
        return

    print(cc.number_jobs)
    print(cc.number_jobs_pending)
    print(cc.number_jobs_running)
    print(cc.number_jobs_finished)
    print('exit status', cc.jobs[0].exit_status())
    # cc.get_number_jobs_failed()

def _do_global_status(nlast):
    table = clusterrun.get_groups_summary(nlast)
    print(tabulate(table))

def do_status(args):
    if args.runid:
        runid = util.proper_runid(args.runid)
        _do_run_status(runid)
    else:
        _do_global_status(args.nlast)

def do_killall(_):
    util.killall(force=True)

# def do_remove_group(args):
#
#     # grps = cluster.get_group_names()
#     if args.what == 'alien':
#         cluster.remove_alien_groups()
#         # rexp = r'/cluster/\d\d\d\d-\d\d-\d\d-\d\d-\d\d-\d\d'
#         # c = re.compile(rexp)
#         # conforming = [g for g in grps if c.match(g)]
#
#
# def _do_remove_run(runids):
#     paths = [join(cluster.base_folder, r) for r in runids]
#     lu.path_.rrm(paths)
#
# def do_remove_run(args):
#     import dateparser
#
#     if args.older_than is None:
#         return
#
#     runids = cluster.get_runids()
#
#     old_runids = []
#     for runid in runids:
#         if dateparser.parse(runid) < dateparser.parse(args.older_than):
#             old_runids.append(runid)
#
#     with lu.BeginEnd('Removing %d folders' % len(old_runids)):
#         _do_remove_run(old_runids)

def entry_point():
    logging.basicConfig()
    p = ArgumentParser()
    sub = p.add_subparsers()

    s = sub.add_parser('root')
    s.set_defaults(func=do_root)

    # s = sub.add_parser('hist')
    # s.add_argument('--last_n', type=int, default=5)
    # s.set_defaults(func=do_hist)

    s = sub.add_parser('status')
    s.add_argument('runid', nargs='?', default=None)
    s.add_argument('--nlast', default=10, type=int)
    s.set_defaults(func=do_status)

    s = sub.add_parser('killall')
    s.set_defaults(func=do_killall)

    #
    # s = sub.add_parser('rm-run')
    # s.add_argument('--older_than', default=None)
    # s.set_defaults(func=do_remove_run)
    #
    # s = sub.add_parser('rm-group')
    # s.add_argument('what', choices=['alien'])
    # s.set_defaults(func=do_remove_group)
    #
    # # s = sub.add_parser('root')
    # # s.set_defaults(func=do_root)

    args = p.parse_args()
    func = args.func
    del args.func
    func(args)
